"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.descriptionFromSpecFile = exports.linePairInsideLinesChanged = exports.lineEndFromNode = exports.lineStartFromNode = void 0;
const ts = require("typescript");
;
let lineStartFromNode = function (file, descNode) {
    return (file.slice(0, descNode['pos'] + 1).split('\n') || []).length + 1;
};
exports.lineStartFromNode = lineStartFromNode;
let lineEndFromNode = function (file, descNode) {
    return (file.slice(0, descNode['end']).split('\n') || []).length;
};
exports.lineEndFromNode = lineEndFromNode;
const linePairInsideLinesChanged = function (subjectLinePair, linesChanged) {
    return linesChanged.some((changedLinePair) => {
        return linePairInRange(subjectLinePair, changedLinePair) || linePairContainsRange(subjectLinePair, changedLinePair);
    });
};
exports.linePairInsideLinesChanged = linePairInsideLinesChanged;
const linePairInRange = function (subjectLinePair, range) {
    const isSubjectStartInChangedRange = subjectLinePair[0] >= range[0] && subjectLinePair[0] <= range[1];
    const isSubjectEndInChangedRange = subjectLinePair[1] >= range[0] && subjectLinePair[1] <= range[1];
    return isSubjectStartInChangedRange || isSubjectEndInChangedRange;
};
const linePairContainsRange = function (subjectLinePair, range) {
    return subjectLinePair[0] <= range[0] && subjectLinePair[1] >= range[1];
};
function docTreeDescribe(node, context, depth, testContext) {
    const lineStart = (0, exports.lineStartFromNode)(context['sourceFile'], node);
    const lineEnd = (0, exports.lineEndFromNode)(context['sourceFile'], node);
    if (context['linesChanged'] !== null && !(0, exports.linePairInsideLinesChanged)([lineStart, lineEnd], context['linesChanged'])) {
        return;
    }
    for (let i = 0; i < depth; i++) {
        context['doc'] = context['doc'].concat(' ');
    }
    const text = node.arguments[0].text;
    context['doc'] = context['doc'].concat(text + '\n');
    const children = descriptionFromNode(node.arguments[1], context, depth + 1, `${testContext}${text}\n`);
    if (children === undefined || children.length === 0) {
        return [];
    }
    return [{
            text: node.arguments[0].text,
            pos: node.pos,
            end: node.end,
            lineStart,
            lineEnd,
            children,
            isAssertion: false,
            isDirectlyModified: true
        }];
}
function docTreeAssertion(node, context, depth, testContext) {
    const lineStart = (0, exports.lineStartFromNode)(context['sourceFile'], node);
    const lineEnd = (0, exports.lineEndFromNode)(context['sourceFile'], node);
    if (context['linesChanged'] !== null && !(0, exports.linePairInsideLinesChanged)([lineStart, lineEnd], context['linesChanged'])) {
        return;
    }
    for (let i = 0; i < depth; i++) {
        context['doc'] = context['doc'].concat(' ');
    }
    const text = node.arguments[0].text;
    context['doc'] = context['doc'].concat(`${node.expression.escapedText} ${text}\n`);
    context['flatDoc'] = `${context['flatDoc']}\n${testContext}${node.expression.escapedText} ${text}\n`;
    return [{
            text,
            pos: node.pos,
            end: node.end,
            lineStart,
            lineEnd,
            children: [],
            isAssertion: true,
            isDirectlyModified: true
        }];
}
function docTree(node, context, depth, testContext) {
    if (node.expression && node.expression.kind == ts.SyntaxKind.Identifier && node.expression.escapedText === 'describe') {
        return docTreeDescribe(node, context, depth, testContext);
    }
    else if (node.expression && node.expression.kind == ts.SyntaxKind.Identifier && (node.expression.escapedText === 'it' || node.expression.escapedText === 'test')) {
        return docTreeAssertion(node, context, depth, testContext);
    }
    else if (node.statements) {
        let nodes = [];
        node.statements.forEach((n) => {
            const nn = descriptionFromNode(n, context, depth + 1, testContext);
            if (nn && nn.length > 0) {
                return nodes.push(...nn);
            }
        });
        if (nodes.length) {
            return nodes;
        }
    }
    else if (node.body) {
        const nn = descriptionFromNode(node.body, context, depth + 1, testContext);
        if (nn && nn.length) {
            return nn;
        }
    }
}
function descriptionFromNode(node, context, depth = 0, testContext = '') {
    let nodes = [];
    ts.forEachChild(node, (n) => {
        const descNode = docTree(n, context, depth, testContext);
        if (descNode) {
            nodes.push(...descNode);
        }
    });
    if (nodes && nodes.length) {
        return nodes;
    }
}
function descriptionFromSpecFile(specFile, linesChanged = null) {
    // linesChanged is an array of line number pairs, e.g. [[1, 3], [5, 5]]
    // an empty array means no lines changed
    // a null value means all lines changed
    let context = { tree: [], doc: '', sourceFile: specFile, linesChanged, flatDoc: '' };
    if (linesChanged !== null && !linesChanged.length) {
        return context;
    }
    const tsSourceFile = ts.createSourceFile('_.js', specFile, ts.ScriptTarget.Latest);
    tsSourceFile.statements.forEach((n) => {
        const nodes = descriptionFromNode(n, context);
        if (nodes) {
            context['tree'].push(...nodes);
        }
    });
    return context;
}
exports.descriptionFromSpecFile = descriptionFromSpecFile;
